from torch.distributions.continuous_bernoulli import ContinuousBernoulli
from torch.nn.modules.loss import _Loss as Loss


class CBLogLoss(Loss):
    """Continuous Bernoulli loss, proposed here:

    https://arxiv.org/abs/1907.06845.
    """

    def __init__(self, reduction="mean", mode="probs"):
        super(CBLogLoss, self).__init__(None, None, reduction)
        self.mode = mode

    def forward(self, input, target):

        # the trick with the dictionary allows us to use either `probs` or `logits`
        log_probs = ContinuousBernoulli(**{self.mode: input}).log_prob(target)

        # sum per input-element log loss
        log_probs = log_probs.view(log_probs.shape[0], -1).sum(axis=1)

        # reduce across batch dimension
        if self.reduction == "none":
            return -log_probs
        elif self.reduction == "sum":
            return -log_probs.sum()
        elif self.reduction == "mean":
            return -log_probs.mean()
        else:
            raise NotImplementedError(f"Unavailable reduction type: {self.reduction}")



import torch


class Project:
    def __init__(self, axis, mode="max"):
        self.axis = axis
        self.mode = mode

    def __call__(self, img):
        if not isinstance(img, torch.Tensor):
            img = torch.tensor(img)

        axis = {"z": -3, "y": -2, "x": -1}

        if self.axis == "z":
            assert len(img.shape) >= 3

        if self.mode == "max":
            return img.max(axis=axis[self.axis])[0]
        elif self.mode == "mean":
            return img.mean(axis=axis[self.axis])
        elif self.mode == "median":
            return img.median(axis=axis[self.axis])
        else:
            raise NotImplementedError



from typing import Sequence

import numpy as np

from .abstract_loader import Loader


class LoadRandomTensor(Loader):
    """Loader class, used to load a dummy tensor, generated randomly.

    It ignores the underlying dataframe
    """

    def __init__(
        self,
        column: str,
        dims: Sequence[int],
    ):
        """
        Parameters
        ----------
        column: str
            Ignored. Kept for consistency with other loaders

        dims:
            Tensor dimensions
        """

        super().__init__()
        self.column = column
        self.dims = dims

    def __call__(self, row):
        return np.random.randn(*self.dims)



import collections
import importlib
import inspect
import json
import os
import yaml
from omegaconf import OmegaConf
from omegaconf._utils import split_key
from .resolvers import _encrypt as encrypt  # noqa: F401


def get_obj_from_path(obj_path: str):
    """Given a class/function/variable path as a string (e.g.
    some.module.ClassName), retrieve the class/function/variable."""
    obj_path = obj_path.split(".")
    module = ".".join(obj_path[:-1])
    name = obj_path[-1]
    return getattr(importlib.import_module(module), name)


def _create_config(obj, obj_path=None, add_partial=False):

    if obj_path is None:
        if hasattr(obj, "__name__"):
            obj_path = obj.__module__ + "." + obj.__name__
        else:
            obj_path = obj.__module__ + "." + obj.__class__.__name__

    if hasattr(obj, "__init__"):
        sig = inspect.getfullargspec(obj.__init__)
    else:
        sig = inspect.getfullargspec(obj)

    args_dict = dict()
    if add_partial:
        args_dict["_partial_"] = True
    args_dict["_target_"] = obj_path

    args = sig.args
    if sig.defaults is not None:
        while len(args) > len(sig.defaults):
            args_dict[args.pop(0)] = None

        assert len(args) == len(sig.defaults)
        for arg, default_value in zip(args, sig.defaults):
            args_dict[arg] = default_value

    kwargs = sig.kwonlyargs
    if sig.kwonlydefaults is not None:
        while len(kwargs) > len(sig.kwonlydefaults):
            args_dict[kwargs.pop(0)] = None

        assert len(args) == len(sig.defaults)
        for arg, default_value in zip(kwargs, sig.kwonlydefaults):
            args_dict[arg] = default_value

    for arg in sig.args + sig.kwonlyargs:
        if arg not in args_dict:
            args_dict[arg] = None

    if "self" in args_dict:
        del args_dict["self"]

    for arg, default in args_dict.items():
        if isinstance(default, tuple):
            args_dict[arg] = list(default)
        elif OmegaConf.is_config(default):
            try:
                args_dict[arg] = _create_config(
                    default, add_partial=inspect.isclass(default)
                )
            except Exception:
                args_dict[arg] = f"Unable to give default value of {type(default)}"

    return args_dict


def create_config(obj_path):
    obj = get_obj_from_path(obj_path)
    config = _create_config(obj, obj_path)
    return OmegaConf.to_yaml(OmegaConf.create(config))


def _maybe_int(value):
    return int(value) if value.isdigit() else value


def add_config(obj_path, dest):
    dest_key = None
    if ":" in dest:
        assert len([c for c in dest if c == ":"]) == 1
        dest, dest_key = dest.split(":")

    obj = get_obj_from_path(obj_path)
    config = _create_config(obj, obj_path)

    if dest_key is not None:
        dest_config = OmegaConf.load(dest)
        dest_path = split_key(dest_key)
        for step in dest_path[::-1]:
            config = {step: config}

        dest_config = OmegaConf.merge(dest_config, config)

    else:
        dest_config = config

    prompt_str_suffix = "" if dest_key is None else f":{dest_key}"
    answer = input(
        f"This is going to overwrite {dest}{prompt_str_suffix}" "\nContinue? [Y/n] "
    )
    if answer is None or answer.lower().strip() in ["", "yes", "y"]:
        with open(dest, "w") as f:
            f.write(OmegaConf.to_yaml(dest_config))
    else:
        print("Not writing the result. The resulting config would have been:\n")
        print(OmegaConf.to_yaml(dest_config))


def deep_merge(template, merge):
    """Recursive dict merge, combines values that are lists.

    This mutates template - the contents of merge are added to template
    (which is also returned). If you want to keep template you could call it
    like deep_merge(dict(template), merge)
    """
    for k, v in merge.items():
        if (
            k in template
            and isinstance(template[k], dict)
            and isinstance(merge[k], collections.Mapping)
        ):
            deep_merge(template[k], merge[k])
        # Removed option to extend the list if there are existing elements:
        # elif k in template and isinstance(template[k], list) and isinstance(v, list):
        #     template[k].extend(v)
        else:
            template[k] = merge[k]
    return template


def merge_config(
    template_path="",
    merge="",
    output_path="",
):
    template_extension = os.path.splitext(template_path)[-1]
    with open(template_path, "r") as template_file:
        if template_extension == ".yaml":
            template = yaml.load(template_file)
        elif template_extension == ".json":
            template = json.load(template_file)
        else:
            raise Exception(f"format {template_extension} not supported")

    output_extension = os.path.splitext(output_path)[-1]
    with open(output_path, "w") as output_file:
        updated = deep_merge(template, merge)

        if output_extension == ".json":
            json.dump(updated, output_file, indent=4)
        elif output_extension == ".yaml":
            yaml.dump(updated, output_file)
        else:
            raise Exception(f"format {output_extension} not supported")



import torch
import torch.nn as nn
import numpy as np

from .abstract_prior import Prior


def compute_tc_penalty(logvar):
    return (2 * logvar).exp().mean(dim=0).sum()


class IsotropicGaussianPrior(Prior):
    def __init__(self, dimensionality=None, tc_penalty_weight=None):
        self.tc_penalty_weight = tc_penalty_weight
        super().__init__(dimensionality)

    @classmethod
    def kl_divergence(cls, mean, logvar, tc_penalty_weight=None, reduction="sum"):
        """Computes the Kullback-Leibler divergence between a diagonal gaussian
        and an isotropic (0,1) gaussian. It also works batch-wise.
        Parameters
        ----------
        mean: torch.Tensor
            Mean of the gaussian (or batch of gaussians)
        logvar: torch.Tensor
            Log-variance of the gaussian (or batch of gaussians)
        """

        kl = -0.5 * (1 + logvar - mean.pow(2) - logvar.exp())

        if tc_penalty_weight is not None:
            tc_penalty = compute_tc_penalty(logvar)
        else:
            tc_penalty = 0

        if reduction == "none":
            return kl
        elif reduction == "mean":
            return kl.mean(dim=-1).mean() + tc_penalty * tc_penalty_weight
        else:
            return kl.sum(dim=-1).mean() + tc_penalty * tc_penalty_weight

    @classmethod
    def sample(cls, mean, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add(mean)

    def forward(self, z, mode="kl", **kwargs):
        mean_logvar = z
        mean, logvar = torch.split(mean_logvar, mean_logvar.shape[1] // 2, dim=1)
        if mode == "kl":
            return self.kl_divergence(
                mean, logvar, tc_penalty_weight=self.tc_penalty_weight, **kwargs
            )
        else:
            return self.sample(mean, logvar, **kwargs)


class DiagonalGaussianPrior(IsotropicGaussianPrior):
    def __init__(
        self,
        dimensionality=None,
        mean=None,
        logvar=None,
        learn_mean=False,
        learn_logvar=False,
        tc_penalty_weight=None,
    ):
        if hasattr(mean, "__len__"):
            if dimensionality is None:
                dimensionality = len(mean)
            else:
                assert dimensionality == len(mean)

        if hasattr(logvar, "__len__"):
            if dimensionality is None:
                dimensionality = len(logvar)
            else:
                assert dimensionality == len(logvar)

        assert dimensionality is not None

        super().__init__(dimensionality)

        if logvar is None:
            logvar = torch.zeros(self.dimensionality)
        else:
            if not hasattr(logvar, "__len__"):
                logvar = [logvar] * self.dimensionality
            logvar = torch.tensor(np.fromiter(logvar, dtype=float))

        if learn_logvar:
            logvar = nn.Parameter(logvar, requires_grad=True)

        self.logvar = logvar

        if mean is None:
            mean = torch.zeros(self.dimension)
        else:
            if not hasattr(mean, "__len__"):
                logvar = [mean] * self.dimensionality
            mean = torch.tensor(np.fromiter(mean, dtype=float))

        if learn_mean:
            mean = nn.Parameter(mean, requires_grad=True)

        self.mean = mean
        self.tc_penalty_weight = tc_penalty_weight

    @classmethod
    def kl_divergence(
        cls, mu1, mu2, logvar1, logvar2, tc_penalty_weight=None, reduction="sum"
    ):
        """Computes the Kullback-Leibler divergence between two diagonal
        gaussians (not necessarily isotropic). It also works batch-wise.
        Parameters
        ----------
        mu1: torch.Tensor
            Mean of the first gaussian (or batch of first gaussians)
        mu2: torch.Tensor
            Mean of the second gaussian (or batch of second gaussians)
        logvar1: torch.Tensor
            Log-variance of the first gaussian (or batch of first gaussians)
        logvar2: torch.Tensor
            Log-variance of the second gaussian (or batch of second gaussians)
        """
        mu_diff = mu2 - mu1

        kl = 0.5 * (
            (logvar2 - logvar1)
            + (logvar1 - logvar2).exp()
            + (mu_diff.pow(2) / logvar2.exp())
            + -1
        )

        if reduction == "none":
            return kl

        loss = kl.sum(dim=-1).mean()

        if tc_penalty_weight is not None:
            tc_penalty = compute_tc_penalty(logvar1)
            loss = loss + tc_penalty_weight * tc_penalty

        return loss

    def forward(self, z, mode="kl", **kwargs):
        mean_logvar = z
        mean, logvar = torch.split(mean_logvar, mean_logvar.shape[1] // 2, dim=1)

        if mode == "kl":
            return self.kl_divergence(
                mean,
                self.mean,
                logvar,
                self.logvar,
                tc_penalty_weight=self.tc_penalty_weight,
                **kwargs
            )

        return self.sample(mean, logvar, **kwargs)



import os
from typing import Optional
import numpy as np

from serotiny.io.image import image_loader
from .abstract_loader import Loader


_DEFAULT_LOADER_KWARGS = dict(
    select_channels=None,
    transform=None,
    reader=None,
    dtype=None,
    return_as_torch=True,
    force_3d=False,
    ome_zarr_level=0,
    ome_zarr_image_name="default",
)


class LoadImage(Loader):
    """Loader class, used to retrieve images from paths given in a dataframe
    column."""

    def __init__(
        self,
        column: str,
        use_cache: bool = False,
        dtype: Optional[str] = None,
        **loader_kwargs,
    ):
        """
        Parameters
        ----------
        column: str
            Dataframe column which contains the image path

        file_type: str
            File format of the image. For now, "tiff" is the
            only supported format. But in the future other formats (like zarr)
            might be supported too

        use_cache: bool = True
            Whether to cache images after downloading them once

        **loader_kwargs
            Additional kwargs passed to `serotiny.io.image.image_loader`

        """
        super().__init__()
        self.column = column

        self.dtype = np.dtype(dtype).type if dtype is not None else None

        self.use_cache = use_cache

        self.loader_kwargs = _DEFAULT_LOADER_KWARGS.copy()
        self.loader_kwargs.update(loader_kwargs)

    def _get_cached_path(self, path):
        if not self.use_cache:
            return path

        conf_path = os.getenv("FSSPEC_CONFIG_DIR")
        if conf_path is not None:
            if "simplecache::" not in str(path):
                return "simplecache::" + path
        return path

    def __call__(self, row):
        path = self._get_cached_path(row[self.column])
        img = image_loader(path, **self.loader_kwargs)
        if self.dtype is not None:
            return self.dtype(img)
        return img



import os
import uuid
import base64
import hashlib
from omegaconf import OmegaConf
from Crypto.Cipher import AES
from Crypto import Random
from getpass import getpass


BLOCK_SIZE = 16


def _pad(s):
    pad_size = BLOCK_SIZE - len(s) % BLOCK_SIZE
    pad_char = chr(pad_size)
    return s + pad_size * pad_char


def _unpad(s):
    return s[: -ord(s[len(s) - 1 :])]


def _get_priv_key():
    password = os.getenv("SEROTINY_PASSWORD")
    if password is None:
        password = getpass("Enter encryption password: ")

    return hashlib.sha256(password.encode("utf-8")).digest()


def _encrypt(raw):
    private_key = _get_priv_key()
    raw = _pad(raw)
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return base64.b64encode(iv + cipher.encrypt(raw.encode())).decode()


def _decrypt(enc):
    private_key = _get_priv_key()
    enc = base64.b64decode(enc)
    iv = enc[:16]
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return _unpad(cipher.decrypt(enc[16:])).decode()


OmegaConf.register_new_resolver("uuid", lambda: str(uuid.uuid4()))
OmegaConf.register_new_resolver("decrypt", _decrypt)



from .abstract_prior import Prior


class IdentityPrior(Prior):
    """Prior class that doesn't contribute to KL loss.

    Effectively a Dirac delta distribution given z.
    """

    def forward(self, z, mode="kl", **kwargs):
        if mode == "kl":
            return 0
        else:
            return z
